/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.impl.javac;


import java.io.IOException;
import java.io.StringReader;
import java.util.LinkedList;
import java.util.List;

import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;
import javax.lang.model.type.TypeKind;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;

import org.crazyjavahacking.astvisualizer.api.ASTParser;
import org.crazyjavahacking.astvisualizer.api.ASTParsingException;
import org.crazyjavahacking.astvisualizer.api.model.EdgeModel;
import org.crazyjavahacking.astvisualizer.api.model.NodeModel;
import org.crazyjavahacking.astvisualizer.api.model.PositionWithinSourceCode;
import org.crazyjavahacking.astvisualizer.impl.javac.model.IComSunSourceTreeTreeEdgeData;
import org.crazyjavahacking.astvisualizer.impl.javac.model.IEdgeData;
import org.crazyjavahacking.astvisualizer.impl.javac.model.IJavaxLangModelElementModifierEdgeData;
import org.crazyjavahacking.astvisualizer.impl.javac.model.IJavaxLangModelElementNameEdgeData;
import org.crazyjavahacking.astvisualizer.impl.javac.model.IJavaxLangModelTypeTypeKindEdgeData;
import org.crazyjavahacking.astvisualizer.impl.javac.util.CollectionUtils;
import org.crazyjavahacking.astvisualizer.impl.javac.util.CompilerTreeAPIUtils;
import org.crazyjavahacking.astvisualizer.impl.javac.util.JavaCompilerUtils;
import org.crazyjavahacking.astvisualizer.util.JavaFileObjectUtils;
import org.jgrapht.DirectedGraph;

import com.sun.source.tree.ClassTree;
import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.ModifiersTree;
import com.sun.source.tree.Tree;
import com.sun.source.util.SourcePositions;
import com.sun.source.util.TreePathScanner;
import com.sun.source.util.Trees;
import com.sun.tools.javac.tree.JCTree;


/**
 * Javac JDK 7 implementation of ASTParser.
 *
 * @author Martin
 */
public final class Javac7ASTParser implements ASTParser {

    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * {@inheritDoc}
     */
    @Override
    public DirectedGraph<NodeModel, EdgeModel> parse(JavaFileObject sourceJavaFileObject) {
        try {
            // first the compilation is required and the source code after compilation will in fact be used later
            // the reason is that after the compilation, the source code will contain syntactic sugar produced by javac
            // compiler (ARM, multidimensional array initializers, ...)
            String sourceCodeAfterCompilation = sourceCodeAfterCompilation(sourceJavaFileObject);
            JavaFileObject javaFileObjectContainingSyntacticSugar =
                    JavaFileObjectUtils.fromReader(new StringReader(sourceCodeAfterCompilation));

            JavacTreePathScanner javacTreePathScanner = new JavacTreePathScanner();

            JavaCompiler.CompilationTask compilationTask =
                    JavaCompilerUtils.getCompilationTaskFromJavaFileObject(javaFileObjectContainingSyntacticSugar);

            CompilationUnitTree compilationUnitTree = JavaCompilerUtils.compile(compilationTask);

            return javacTreePathScanner.doScanning(compilationTask, compilationUnitTree);
        }
        catch (Exception e) {
            throw new ASTParsingException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String compilerName() {
        return "javac";
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String compilerVersion() {
        return "1.7";
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Private Helpers.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Returns source code after compilation.
     *
     * @param javaFileObject JavaFileObject
     *
     * @return source code after compilation
     */
    private String sourceCodeAfterCompilation(JavaFileObject javaFileObject) {
        CompilationUnitTree compilationUnitTree = JavaCompilerUtils.compile(javaFileObject);

        return compilationUnitTree.toString();
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Inner classes.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Javac tree path scanner.
     */
    private static final class JavacTreePathScanner extends TreePathScanner<Integer, Void> {

        //-------------------------------------------------------------------------------------------------------------
        // Constants.
        //-------------------------------------------------------------------------------------------------------------
        private static final String CONSTRUCTOR_METHOD_NAME = "<init>";


        //-------------------------------------------------------------------------------------------------------------
        // Members.
        //-------------------------------------------------------------------------------------------------------------
        private final LinkedList<NodeModel> nodeModelStack = new LinkedList<NodeModel>();
        private final LinkedList<IEdgeData> edgeDataStack  = new LinkedList<IEdgeData>();

        private final CustomizedAST customizedAST = new CustomizedAST();

        private SourcePositionHelper sourcePositionHelper = null;

        private String lastSimpleClassName = null;


        //-------------------------------------------------------------------------------------------------------------
        // Private Helpers.
        //-------------------------------------------------------------------------------------------------------------
        /**
         * Starts scanning of given compilation unit tree as part of given compilation task.
         *
         * @param compilationTask     compilation task
         * @param compilationUnitTree compilation unit tree
         *
         * @throws IllegalStateException if either of nodeModelStack or edgeDataStack is empty
         *
         * @return customized AST
         */
        private DirectedGraph<NodeModel, EdgeModel> doScanning(JavaCompiler.CompilationTask compilationTask,
                CompilationUnitTree compilationUnitTree) {

            SourcePositions sourcePositions = Trees.instance(compilationTask).getSourcePositions();
            sourcePositionHelper = new SourcePositionHelper(sourcePositions, compilationUnitTree, customizedAST);

            scan(compilationUnitTree, (Void) null);

            connectRemainingNodes();

            if (!nodeModelStack.isEmpty()) {
                throw new IllegalStateException("Node model stack is not empty ! : " + nodeModelStack);
            }

            if (!edgeDataStack.isEmpty()) {
                throw new IllegalStateException("Edge data stack is not empty ! : " + edgeDataStack);
            }

            return customizedAST.toDirectedGraph();
        }

        /**
        * {@inheritDoc}
        */
        @Override
        public Integer scan(Tree tree, Void p) {
            if (tree != null) {
                if (tree.getKind() == Tree.Kind.CLASS) {
                    lastSimpleClassName = ((ClassTree) tree).getSimpleName().toString();
                }

                processTreeAndAddToCustomizedAST(tree);
            }

            return super.scan(tree, p);
        }

        /**
        * Processes given Tree instance and adds it to customized AST.
        *
        * @param treeNode tree node
        */
        private void processTreeAndAddToCustomizedAST(Tree treeNode) {
            if (CompilerTreeAPIUtils.isNotCompilationUnitTree(treeNode)) {
                processSpecialElements();
            }

            // !!! special treatment of empty ModifiersTree (without flags and annotations)
            // -> this tree will not be added even if it is part of the AST produced by javac
            if (CompilerTreeAPIUtils.isModifiersTree(treeNode)) {
                ModifiersTree modifiersTree = (ModifiersTree) treeNode;

                if (!ModifiersTreeHelper.hasFlagsOrAnnotations(modifiersTree)) {
                    return;
                }
            }

            addNodeAndEdgeToCustomizedAST(treeNode);
        }

        /**
         * Processes special elements:
         * <ul>
         *     <li>javax.lang.model.element.Name</li>
         *     <li>javax.lang.model.element.Modifier</li>
         *     <li>javax.lang.model.type.TypeKind</li>
         * </ul>
         */
        private void processSpecialElements() {
            IEdgeData topEdgeIdentity = edgeDataStack.peek();

            if (topEdgeIdentity == null) {
                // no more data to process
                return;
            }
            if (topEdgeIdentity instanceof IComSunSourceTreeTreeEdgeData) {
                // com.sun.source.tree.Tree is not special element
                return;
            }

            NodeModel nodeModel = nodeModelFromSpecialElement(topEdgeIdentity);
            customizedAST.addNode(nodeModel);

            String transitionName = edgeDataStack.pop().getTransitionName();
            EdgeModel edge = EdgeModelFactory.fromTransitionName(transitionName);

            customizedAST.addEdge(nodeModelStack.pop(), // edge source
                                  nodeModel,            // edge target
                                  edge);

            processSpecialElements(); // recursion...
        }

        /**
         * Creates NodeModel from special element.
         *
         * @param iEdgeData IEdgeData
         *
         * @see org.crazyjavahacking.astvisualizer.impl.javac.Javac7ASTParser.JavacTreePathScanner#processSpecialElements()
         *         processSpecialElements()
         *
         * @return NodeModel
         */
        private NodeModel nodeModelFromSpecialElement(IEdgeData iEdgeData) {
            int currentDepth = nodeModelStack.peek().getDepth() + 1;
            int nodeCounter = customizedAST.getNumberOfNodes() + 1;

            NodeModel nodeModel;

            if (iEdgeData instanceof IJavaxLangModelElementNameEdgeData) {
                Name name = ((IJavaxLangModelElementNameEdgeData) iEdgeData).getName();

                String textRepresentation = name.toString();
                if (textRepresentation.equals(CONSTRUCTOR_METHOD_NAME)) {
                    textRepresentation = lastSimpleClassName;
                }

                nodeModel = NodeModelFactory.fromJavaxLangModelElementName(
                                                        currentDepth,
                                                        nodeCounter,
                                                        name,
                                                        sourcePositionHelper.determinePositionFromText(textRepresentation));
            }
            else if (iEdgeData instanceof IJavaxLangModelElementModifierEdgeData) {
                Modifier modifier = ((IJavaxLangModelElementModifierEdgeData) iEdgeData).getModifier();
                nodeModel = NodeModelFactory.fromJavaxLangModelElementModifier(
                                                    currentDepth,
                                                    nodeCounter,
                                                    modifier,
                                                    sourcePositionHelper.determinePositionFromText(modifier.toString()));
            }
            else if (iEdgeData instanceof IJavaxLangModelTypeTypeKindEdgeData) {
                TypeKind typeKind = ((IJavaxLangModelTypeTypeKindEdgeData) iEdgeData).getTypeKind();
                nodeModel = NodeModelFactory.fromJavaxLangModelTypeTypeKind(
                                                    currentDepth,
                                                    nodeCounter,
                                                    typeKind,
                                                    sourcePositionHelper.determinePositionFromText(typeKind.toString()
                                                                                                       .toLowerCase()));
            }
            else {
                throw new IllegalArgumentException(
                        "Given IEdgeData implementation is not special element! : " + iEdgeData.getClass());
            }

            return nodeModel;
        }

        /**
         * Adds node and edge to customzed AST.
         *
         * @param treeNode tree node
         */
        private void addNodeAndEdgeToCustomizedAST(Tree treeNode) {
            int currentDepth =
                    (CompilerTreeAPIUtils.isCompilationUnitTree(treeNode)) ? 0
                                                                           : nodeModelStack.peek().getDepth() + 1;
            int nodeCounter = customizedAST.getNumberOfNodes() + 1;

            PositionWithinSourceCode positionWithinSourceCode = sourcePositionHelper.determinePositionFromTree(treeNode);
            String textRepresentation = sourcePositionHelper.determineTextFromPosition(positionWithinSourceCode);

            NodeModel nodeModel = NodeModelFactory.fromComSourceSunTreeTree(currentDepth,
                                                                            nodeCounter,
                                                                            textRepresentation,
                                                                            treeNode,
                                                                            positionWithinSourceCode);

            customizedAST.addNode(nodeModel);

            if (CompilerTreeAPIUtils.isNotCompilationUnitTree(treeNode)) {
                if (!edgeDataStack.isEmpty()) {
                    String transitionName = edgeDataStack.pop().getTransitionName();
                    EdgeModel edge = EdgeModelFactory.fromTransitionName(transitionName);

                    customizedAST.addEdge(nodeModelStack.pop(), // edge source
                                          nodeModel,            // edge target
                                          edge);
                }
            }

            fillNodeModelAndEdgeDataStacks(treeNode, nodeModel);
        }

        /**
        * Fills node model and edge data stacks.
        *
        * @param treeNode  tree node
        * @param nodeModel node model
        */
        private void fillNodeModelAndEdgeDataStacks(Tree treeNode, NodeModel nodeModel) {
            @SuppressWarnings("unchecked")
            Class<? extends JCTree> treeNodeType = (Class<? extends JCTree>) (Class<?>) treeNode.getClass();

            TreeTypeEnum treeType = TreeTypeEnum.getByClass(treeNodeType);
            List<IEdgeData> edgeDatas = treeType.getEdgeData(treeNode);

            for (int count = 0; count < edgeDatas.size(); count++) {
                nodeModelStack.push(nodeModel); // order does not matter, all objects are in fact the same
            }

            List<IEdgeData> edgeDataReversedList = CollectionUtils.reverseListOrder(edgeDatas);
            for (IEdgeData edgeData : edgeDataReversedList) {
                edgeDataStack.push(edgeData); // order matter, it is stack, so first collection was reversed
            }
        }

        /**
         * Connects remaining nodes.
         */
        private void connectRemainingNodes() {
            if (nodeModelStack.isEmpty() && edgeDataStack.isEmpty()) {
                // nothing to connect
                return;
            }

            processSpecialElements();

            if (nodeModelStack.isEmpty() && edgeDataStack.isEmpty()) {
                // nothing to connect - all remaning nodes + edges were connected as part of processSpecialElemenets()
                // method call
                return;
            }

            IEdgeData topEdgeIdentity = edgeDataStack.peek();
            if (topEdgeIdentity == null) {
                // no more data to process
                return;
            }
            if (topEdgeIdentity instanceof IComSunSourceTreeTreeEdgeData) {
                throw new AssertionError();
            }

            NodeModel nodeModel = nodeModelFromSpecialElement(topEdgeIdentity);
            customizedAST.addNode(nodeModel);

            if (!edgeDataStack.isEmpty()) {
                String transitionName = edgeDataStack.pop().getTransitionName();
                EdgeModel upperEdge = EdgeModelFactory.fromTransitionName(transitionName);

                customizedAST.addEdge(nodeModelStack.pop(), // edge source
                                      nodeModel,            // edge target
                                      upperEdge);
            }

            if (!nodeModelStack.isEmpty() && !edgeDataStack.isEmpty()) {
                connectRemainingNodes(); // recursion
            }
        }
    }

    /**
     * Source positions helper.
     */
    private static final class SourcePositionHelper {

        //-------------------------------------------------------------------------------------------------------------
        // Members.
        //-------------------------------------------------------------------------------------------------------------
        private final SourcePositions sourcePositions;
        private final CompilationUnitTree compilationUnitTree;
        private final CustomizedAST customizedAST;

        private final String wholeSourceText;


        //-------------------------------------------------------------------------------------------------------------
        // Costructors.
        //-------------------------------------------------------------------------------------------------------------
        private SourcePositionHelper(SourcePositions sourcePositions, CompilationUnitTree compilationUnitTree,
                CustomizedAST customizedAST) {

            this.sourcePositions = sourcePositions;
            this.compilationUnitTree = compilationUnitTree;
            this.customizedAST = customizedAST;

            try {
                this.wholeSourceText = compilationUnitTree.getSourceFile()
                                                          .getCharContent(false)
                                                          .toString();
            }
            catch (IOException e) {
                throw new IllegalStateException("IOException during getting JavaFileObject content...");
            }
        }


        //-------------------------------------------------------------------------------------------------------------
        // Private Helpers.
        //-------------------------------------------------------------------------------------------------------------
        /**
         * Determines source position from given tree.
         *
         * @param tree source tree
         *
         * @return tree source position
         */
        private PositionWithinSourceCode determinePositionFromTree(Tree tree) {
            long startSourcePosition = sourcePositions.getStartPosition(compilationUnitTree, tree);
            long endSourcePosition   = sourcePositions.getEndPosition  (compilationUnitTree, tree);

            return new PositionWithinSourceCode(startSourcePosition, // inclusive
                                                endSourcePosition);  // exclusive
        }

        /**
         * Determines source position from given text according to last added node to Customized AST.
         *
         * @param text text
         *
         * @return text source position
         */
        private PositionWithinSourceCode determinePositionFromText(String text) {
            int fromIndexIncluding;

            NodeModel previousNodeModel = customizedAST.getLastAddedNode();
            if (previousNodeModel.isLeaf()) {
                // |+1| is needed because there could be special case of parsing "private e" and e must not be found as
                //      part of the "private"
                fromIndexIncluding = (int) previousNodeModel.getPositionWithinSourceCode().getEnd() + 1;
            }
            else {
                fromIndexIncluding = (int) previousNodeModel.getPositionWithinSourceCode().getStart();
            }

            long startPosition = wholeSourceText.indexOf(text,
                                                         fromIndexIncluding); // inclusive
            long endPosition = startPosition + text.length();                 // exclusive

            return new PositionWithinSourceCode(startPosition, endPosition);
        }

        private String determineTextFromPosition(PositionWithinSourceCode sourcePosition) {
            return wholeSourceText.substring((int) sourcePosition.getStart(),
                                             (int) sourcePosition.getEnd());
        }
    }
}
